import functools
import requests

from requests.adapters import HTTPAdapter, Retry


class ModelSessionMeta(type):
    """Meta class that is used to create a static attribute called request.
    This request attribute can be used just like the regular 'requests'
    function from the requests module.

    This prevents having to do a new handshake for every request
    we send towards the API.
    """

    @property
    def request(cls) -> requests.Session:
        if not cls._request_session:
            cls._request_session = requests.Session()
            cls._request_session.request = functools.partial(
                cls._request_session.request,
                timeout=9,
            )
            adapter = HTTPAdapter(
                max_retries=Retry(
                    total=2,
                    backoff_factor=0.3,
                    status_forcelist=[500, 502, 503, 504],
                )
            )
            cls._request_session.mount("http://", adapter)
            cls._request_session.mount("https://", adapter)
        return cls._request_session


class ModelSessionMixin(metaclass=ModelSessionMeta):
    _request_session: requests.Session | None = None



from typing import Generator, Type, TypeVar

import uweb3
from uweb3.libs.sqltalk.mysql.connection import Connection

T = TypeVar("T", bound="RichModel")


class RichModel(uweb3.model.Record):
    """Provides a richer uweb Record class."""

    SEARCHABLE_COLUMNS = []

    def PagedChildren(self, classname, *args, **kwargs):
        """Return child objects with extra argument options."""
        if "conditions" in kwargs:
            kwargs["conditions"].append("%s = %d" % (self.TableName(), self.key))
        else:
            kwargs["conditions"] = "%s = %d" % (self.TableName(), self.key)
        if "offset" in kwargs and kwargs["offset"] < 0:
            kwargs["offset"] = 0
        return classname.List(*args, **kwargs)

    @classmethod
    def List(
        cls: Type[T],
        connection: Connection,
        conditions=None,
        limit=None,
        offset=None,
        order=None,
        yield_unlimited_total_first=False,
        search=None,
        tables=None,
        escape=True,
        fields=None,
    ) -> Generator[T, None, None]:
        """Yields a Record object for every table entry.

        Arguments:
          @ connection: object
            Database connection to use.
          % conditions: str / iterable ~~ None
            Optional query portion that will be used to limit the list of results.
            If multiple conditions are provided, they are joined on an 'AND' string.
          % limit: int ~~ None
            Specifies a maximum number of items to be yielded. The limit happens on
            the database side, limiting the query results.
          % offset: int ~~ None
            Specifies the offset at which the yielded items should start. Combined
            with limit this enables proper pagination.
          % order: iterable of str/2-tuple
            Defines the fields on which the output should be ordered. This should
            be a list of strings or 2-tuples. The string or first item indicates the
            field, the second argument defines descending order (desc. if True).
          % yield_unlimited_total_first: bool ~~ False
            Instead of yielding only Record objects, the first item returned is the
            number of results from the query if it had been executed without limit.
          % search: str
            Specifies what string should be searched for in the default searchable
            database columns.

        Yields:
          Record: Database record abstraction class.
        """
        if not tables:
            tables = [cls.TableName()]
        group = None
        if fields is None:
            fields = "%s.*" % cls.TableName()
        if search:
            search = search.strip()
            group = "%s.%s" % (
                cls.TableName(),
                (
                    cls.RecordKey()
                    if getattr(cls, "RecordKey", None)
                    else cls._PRIMARY_KEY
                ),
            )
            tables, newconditions = cls._GetColumnData(tables, search)
            if conditions:
                if type(conditions) == list:
                    conditions.extend(newconditions)
                else:
                    newconditions.append(conditions)
                    conditions = newconditions
            else:
                conditions = newconditions
        with connection as cursor:
            if hasattr(cls, "_addToCache"):
                connection.modelcache["_stats"]["queries"].append(
                    "%s VersionedRecord.List" % cls.TableName()
                )
            records = cursor.Select(
                fields=fields,
                table=tables,
                conditions=conditions,
                limit=limit,
                offset=offset,
                order=order,
                totalcount=yield_unlimited_total_first,
                escape=escape,
                group=group,
            )
        if yield_unlimited_total_first:
            yield records.affected
        records = [cls(connection, record) for record in list(records)]
        for record in records:
            yield record
        if hasattr(cls, "_addToCache"):
            # and not fields or fields == '*':
            # dont cache partial objects
            list(cls._cacheListPreseed(records))

    @classmethod
    def List(
        cls: Type[T],
        connection: Connection,
        conditions=None,
        limit=None,
        offset=None,
        order=None,
        yield_unlimited_total_first=False,
        search=None,
        tables=None,
        escape=True,
        fields=None,
    ) -> Generator[T, None, None]:
        """Yields a Record object for every table entry.

        Arguments:
          @ connection: object
            Database connection to use.
          % conditions: str / iterable ~~ None
            Optional query portion that will be used to limit the list of results.
            If multiple conditions are provided, they are joined on an 'AND' string.
          % limit: int ~~ None
            Specifies a maximum number of items to be yielded. The limit happens on
            the database side, limiting the query results.
          % offset: int ~~ None
            Specifies the offset at which the yielded items should start. Combined
            with limit this enables proper pagination.
          % order: iterable of str/2-tuple
            Defines the fields on which the output should be ordered. This should
            be a list of strings or 2-tuples. The string or first item indicates the
            field, the second argument defines descending order (desc. if True).
          % yield_unlimited_total_first: bool ~~ False
            Instead of yielding only Record objects, the first item returned is the
            number of results from the query if it had been executed without limit.
          % search: str
            Specifies what string should be searched for in the default searchable
            database columns.

        Yields:
          Record: Database record abstraction class.
        """
        if not tables:
            tables = [cls.TableName()]
        group = None
        if fields is None:
            fields = "%s.*" % cls.TableName()
        if search:
            search = search.strip()
            group = "%s.%s" % (
                cls.TableName(),
                (
                    cls.RecordKey()
                    if getattr(cls, "RecordKey", None)
                    else cls._PRIMARY_KEY
                ),
            )
            tables, newconditions = cls._GetColumnData(tables, search)
            if conditions:
                if type(conditions) == list:
                    conditions.extend(newconditions)
                else:
                    newconditions.append(conditions)
                    conditions = newconditions
            else:
                conditions = newconditions
        with connection as cursor:
            if hasattr(cls, "_addToCache"):
                connection.modelcache["_stats"]["queries"].append(
                    "%s VersionedRecord.List" % cls.TableName()
                )
            records = cursor.Select(
                fields=fields,
                table=tables,
                conditions=conditions,
                limit=limit,
                offset=offset,
                order=order,
                totalcount=yield_unlimited_total_first,
                escape=escape,
                group=group,
            )
        if yield_unlimited_total_first:
            yield records.affected
        records = [cls(connection, record) for record in list(records)]
        for record in records:
            yield record
        if hasattr(cls, "_addToCache"):
            # and not fields or fields == '*':
            # dont cache partial objects
            list(cls._cacheListPreseed(records))

    @classmethod
    def _GetColumnData(cls, tables, search):
        """Extracts table information from the searchable columns."""
        conditions = []
        # XXX search needs to be escaped properly
        condition = 'like "%%%s%%" or ' % search
        searchcondition = ""
        for column in cls.SEARCHABLE_COLUMNS:
            columndata = column.split(".")
            if len(columndata) == 2:
                classname = columndata[0]
                table = cls._SUBTYPES[classname]
                fkey = cls._FOREIGN_RELATIONS.get(classname, False)
                if fkey and fkey.get("LookupKey", False):
                    key = fkey.get("LookupKey")
                elif getattr(table, "RecordKey", None):
                    key = table.RecordKey()
                else:
                    key = table._PRIMARY_KEY
                conditions.append(
                    "`%s`.`%s` = %s.%s"
                    % (cls.TableName(), table.TableName(), table.TableName(), key)
                )
                if (
                    table.TableName() not in tables
                    and table.TableName() != cls.TableName()
                ):
                    tables.append(table.TableName())
                searchcondition += "`%s`.`%s` %s" % (
                    table.TableName(),
                    columndata[1],
                    condition,
                )
            else:
                searchcondition += "`%s`.`%s` %s" % (cls.TableName(), column, condition)
        if searchcondition:
            searchcondition = (
                "(%s)" % searchcondition[:-4]
            )  # TODO use ' or '.join on search conditions instead
            conditions.append(searchcondition)
        return tables, conditions



import configparser

import uweb3


class RestrictedDebuggingMixin(uweb3.DebuggingPageMaker):
    """This debugging pagemaker only shows debugging info to those on the white-list"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_development = self.config.config.getboolean("general", "development")

    def _should_show_exception_page(self):
        """Determine whether we should show the exception page to the user or not.

        If the application is run in development mode always show the page.
        If the application is run in production mode only show the page if one
            of the following conditions are met:
                - The users IP address is in the white-list
                - The user is using a forawarded IP address
                    and the forwarded IP address is in the white-list
                    and the application allows forwarded IP addresses
        """
        if self.is_development:
            return True

        if (
            "debugging" not in self.options
            or "whitelist" not in self.options["debugging"]
        ):
            return False

        # Comma seperated list of IP addresses
        whitelist = self.options["debugging"]["whitelist"]
        if not whitelist:
            return False

        # Make sure no empty strings are in the list.
        whitelisted_addresses = list(filter(None, whitelist.split(",")))
        remote_address = self.req.env["REMOTE_ADDR"]

        if remote_address in whitelisted_addresses:
            return True

        # Attempt to parse the value to a boolean, default to false.
        try:
            useforwardedip = self.config.config.getboolean(
                "debugging", "useforwardedip"
            )
        except ValueError:
            self.logger.warning("Invalid useforwardedip setting found, assuming False")
            return False
        except configparser.NoOptionError:
            # Option was not present, so user is not authorized to see the page.
            return False

        try:
            real_address = self.req.env["HTTP_X_FORWARDED_FOR"]
        except KeyError:
            return False

        if useforwardedip and real_address in whitelisted_addresses:
            return True

        return False

    def InternalServerError(self, exc_type, exc_value, traceback):
        self.logger.exception(
            "Internal server error occured: ", exc_info=(exc_type, exc_value, traceback)
        )
        if self._should_show_exception_page():
            return super().InternalServerError(exc_type, exc_value, traceback)
        return self.Error("A server error occurred.", httpcode=500, log_error=False)  # type: ignore # noqa: E501



from __future__ import annotations

import math
from abc import abstractmethod

from uweb3.libs.safestring import HTMLsafestring
from uweb3plugins.core.paginators.columns import Col
from uweb3plugins.core.paginators.html_elements import (
    SearchField,
    Table,
    TableBody,
    TableHeader,
    TablePagination,
)


def get_current_page(get_request_data):
    try:
        return int(get_request_data.getfirst("page", 1))
    except (ValueError, KeyError):
        return 1


def calc_total_pages(total_items: int, items_per_page: int):
    return int(math.ceil(float(total_items) / items_per_page))


class MetaTable(type):
    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)
        cls._columns = {
            name: obj for name, obj in attrs.items() if isinstance(obj, Col)
        }
        return cls


class BasicTable(metaclass=MetaTable):
    def __init__(
        self,
        items,
        sort_by=None,
        sort_direction=None,
        search_url=None,
        page=None,
        total_pages=None,
        renderer: None | "RenderCustomTable" = None,
        query: None | str = None,
    ):
        self.items = items
        self.sort_by = sort_by
        self.search_url = search_url
        self.page = page
        self.total_pages = total_pages
        self.query = query

        if not renderer:
            self.renderer = RenderSimpleTable()
        else:
            self.renderer = renderer

        if self.sort_by and not sort_direction:
            # TODO: Warning?
            self.sort_direction = "ASC"
        else:
            self.sort_direction = sort_direction

    def _get_columns(self):
        yield from [col for col in self._columns.values() if col.enabled]

    @property
    def render(self):
        return self.renderer.render(self)


class TableComponents:
    def __init__(self):
        self._components = []

    def add_component(self, component):
        self._components.append(component)

    def render(self, table: BasicTable):
        return HTMLsafestring("").join(
            component(table=table).render for component in self._components
        )


class RenderCustomTable:
    @abstractmethod
    def __init__(self):
        self._renderer = TableComponents()

    def render(self, table: BasicTable):
        return self._renderer.render(table)


class RenderCompleteTable(RenderCustomTable):
    def __init__(self):
        self._renderer = TableComponents()
        self._renderer.add_component(SearchField)
        self._renderer.add_component(
            Table(
                [
                    TableHeader,
                    TableBody,
                ]
            )
        )
        self._renderer.add_component(TablePagination)


class RenderSimpleTable(RenderCustomTable):
    def __init__(self):
        self._renderer = TableComponents()
        self._renderer.add_component(
            Table(
                [
                    TableHeader,
                    TableBody,
                ]
            )
        )



from uweb3plugins.core.paginators.html_elements import Element
from uweb3plugins.core.paginators import helpers
import string


class MyFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        return (self.get_value(field_name, args, kwargs), field_name)


class Col:
    def __init__(
        self,
        name,
        attr,
        sortable=False,
        enabled=True,
        value_formatter=None,
    ):
        """Base column class that can be used to render simple
                columns in a table.
        u
                Args:
                    name (str): The title of the column, this will be displayed
                        in the <th> element.
                    attr (str): The attribute name of the value that should be
                        retrieved from the item. When the item has nested children
                        (because of lazyloading) or the item is a dict with nested
                        keys the following syntax can be used to extract the value
                        from a nested object: "foo.bar.baz". This will return the
                        value of item["foo"]["bar"]["baz"].
                    sortable (bool, optional): Determines if the column is sortable.
                        Defaults to False.
                    enabled (bool, optional): Determine wheter or not
                        this column header and value should be displayed.
                        Defaults to True.
                    value_formatter (Callable, optional): A callable that is called
                        right before the column value is renderd. The result of
                        the value_formatter will be used as value in the element.

                Example usage:
                    class Items:
                        ID: int
                        name: str

                    class ItemsTable(BasicTable):
                        id = Col("ID", attr="ID")
                        name = Col("Name", attr="name")

                    table = ItemsTable(items=[Items(ID=1, name="foo"), Items(ID=2, name="bar")])

                    The rendered table will look like this:
                        <table>
                            <thead>
                                <tr>
                                    <th>ID</th>
                                    <th>Name</th>
                                </tr>
                            </thead>
                            <tbody>
                                <tr>
                                    <td>1</td>
                                    <td>foo</td>
                                </tr>
                                <tr>
                                    <td>2</td>
                                    <td>bar</td>
                                </tr>
                            </tbody>
                        </table>

        """
        self.name = name
        self.attr = attr
        self.sortable = sortable
        self.enabled = enabled
        self.value_formatter = value_formatter

    def render(self, item):
        value = helpers.get_attr(item, self.attr)
        if self.value_formatter:
            value = self.value_formatter(value)
        return Element("td", value=value)


class LinkCol(Col):
    def __init__(self, name, attr, href, *args, **kwargs):
        """Adds a row to the table of the following format:
            <td>
                <a href="href">attr</a>
            </td>

        The value that is displayed is the passed attr value.
        The column will be populated with data on render, the
        value in the href will be retrieved from the current item
        that is being iterated over.

        Retrival of nested values or children from uweb3 model
        objects is supported. For example if we have a record
        class that looks like this:

            class Client(uweb3.model.Record):
                ID: int
                name: str

            class Invoice(uweb3.model.Record):
                ID: int
                title: str
                client: Client.ID

            class Table(BasicTable):
                col1 = LinkCol("Client", attr="client.name", href="/client/{client.ID}")
                col2 = LinkCol("Invoice", attr="title", href="/invoices/{ID}")

            Table(items=[Invoice(1, "Invoice 1", Client(1, "Client 1"))])

            The rendered table will look like this:
                <table>
                    <thead>
                        <tr>
                            <th>Client</th>
                            <th>Invoice</th>
                        </tr>
                    </thead>
                    <tbody>
                        <tr>
                            <td>
                                <a href="/client/1">Client 1</a>
                            </td>
                            <td>
                                <a href="/invoices/1">Invoice 1</a>
                            </td>
                        </tr>
                    </tbody>
                </table>

        It is also possible to pass an href without any fields:
            class Table(BasicTable):
                LinkCol("Invoice", attr="title", href="/invoices")

            which will result in the following table:
                <table>
                    <thead>
                        <tr>
                            <th>Invoice</th>
                        </tr>
                    </thead>
                    <tbody>
                        <tr>
                            <td>
                                <a href="/invoices">Invoice 1</a>
                            </td>
                        </tr>
                    </tbody>
                </table>
        """
        super().__init__(name, attr, *args, **kwargs)
        self.href = href

    def render(self, item):
        # Retrieve all fields that should be replaced in the href.
        # When no {replacements} are found just use the href as is.
        field_names = [
            name
            for text, name, spec, conv in string.Formatter().parse(self.href)
            if name
        ]
        # This is some fuckery that allows formatting replacements like
        # {replacement.key}, normally this is not allowed because
        # dot syntax is not supported in fstrings. This is a workaround.
        formatted_url = MyFormatter().vformat(
            self.href, [], {name: helpers.get_attr(item, name) for name in field_names}
        )
        value = helpers.get_attr(item, self.attr)
        if self.value_formatter:
            value = self.value_formatter(value)

        return Element(
            "td",
            children=[
                Element(
                    "a",
                    value=value,
                    attrs=f'href="{formatted_url}"',
                )
            ],
        )


class ConstantAttr:
    def __init__(self, value):
        """This class can be used to pass a constant value to a column.
        This supresses the default behaviour of retrieving the value
        from the item that is being iterated over."""
        self.attr = value

    def __repr__(self):
        return str(self.attr)



from typing import Generator, Type, TypeVar

import uweb3
from uweb3.libs.sqltalk.mysql.connection import Connection

T = TypeVar("T", bound="RichVersionedRecord")


class RichVersionedRecord(uweb3.model.VersionedRecord):
    """Provides a richer uweb VersionedRecord class."""

    SEARCHABLE_COLUMNS = []

    @classmethod
    def List(
        cls: Type[T],
        connection: Connection,
        conditions=None,
        limit=None,
        offset=None,
        order=None,
        yield_unlimited_total_first=False,
        search=None,
        tables=None,
        escape=True,
        fields=None,
    ) -> Generator[T, None, None]:
        """Yields the latest Record for each versioned entry in the table.

        Arguments:
        @ connection: object
          Database connection to use.
        % conditions: str / iterable ~~ None
          Optional query portion that will be used to limit the list of results.
          If multiple conditions are provided, they are joined on an 'AND' string.
        % limit: int ~~ None
          Specifies a maximum number of items to be yielded. The limit happens on
          the database side, limiting the query results.
        % offset: int ~~ None
          Specifies the offset at which the yielded items should start. Combined
          with limit this enables proper pagination.
        % order: iterable of str/2-tuple
          Defines the fields on which the output should be ordered. This should
          be a list of strings or 2-tuples. The string or first item indicates the
          field, the second argument defines descending order (desc. if True).
        % yield_unlimited_total_first: bool ~~ False
          Instead of yielding only Record objects, the first item returned is the
          number of results from the query if it had been executed without limit.
        % search: str
          Specifies what string should be searched for in the default searchable
          database columns.

        Yields:
          Record: The Record with the newest version for each versioned entry.
        """
        if not tables:
            tables = [cls.TableName()]
        if not fields:
            fields = "%s.*" % cls.TableName()
        else:
            if fields != "*":
                if type(fields) != str:
                    fields = ", ".join(connection.EscapeField(fields))
                else:
                    fields = connection.EscapeField(fields)
        if search:
            search = search.strip()
            tables, newconditions = cls._GetColumnData(tables, search)
            if conditions:
                if type(conditions) == list:
                    conditions.extend(newconditions)
                else:
                    newconditions.append(conditions)
                    conditions = newconditions
            else:
                conditions = newconditions
        field_escape = connection.EscapeField if escape else lambda x: x
        if yield_unlimited_total_first and limit is not None:
            totalcount = "SQL_CALC_FOUND_ROWS"
        else:
            totalcount = ""
        with connection as cursor:
            records = cursor.Execute(
                """
          SELECT %(totalcount)s %(fields)s
          FROM %(tables)s
          JOIN (SELECT MAX(`%(primary)s`) AS `max`
                FROM `%(table)s`
                GROUP BY `%(record_key)s`) AS `versions`
              ON (`%(table)s`.`%(primary)s` = `versions`.`max`)
          WHERE %(conditions)s
          %(order)s
          %(limit)s
          """
                % {
                    "totalcount": totalcount,
                    "primary": cls._PRIMARY_KEY,
                    "record_key": cls.RecordKey(),
                    "fields": fields,
                    "table": cls.TableName(),
                    "tables": cursor._StringTable(tables, field_escape),
                    "conditions": cursor._StringConditions(conditions, field_escape),
                    "order": cursor._StringOrder(order, field_escape),
                    "limit": cursor._StringLimit(limit, offset),
                }
            )
        if yield_unlimited_total_first and limit is not None:
            with connection as cursor:
                records.affected = cursor._Execute("SELECT FOUND_ROWS()")[0][0]
            yield records.affected
        # turn sqltalk rows into model
        records = [cls(connection, record) for record in list(records)]
        for record in records:
            yield record
        if (
            hasattr(cls, "_addToCache")
            and not fields
            or (fields == "*" and len(tables) == 1)
        ):
            list(cls._cacheListPreseed(records))

    @classmethod
    def _GetColumnData(cls, tables, search):
        """Extracts table information from the searchable columns."""
        conditions = []
        # XXX search needs to be escaped properly
        condition = 'like "%%%s%%" or ' % search
        searchcondition = ""
        for column in cls.SEARCHABLE_COLUMNS:
            columndata = column.split(".")
            if len(columndata) == 2:
                classname = columndata[0][0].upper() + columndata[0][1:]
                table = globals()[classname]
                if getattr(table, "RecordKey", None):
                    key = table.RecordKey()
                else:
                    key = table._PRIMARY_KEY
                conditions.append(
                    "`%s`.`%s` = %s.%s"
                    % (cls.TableName(), table.TableName(), table.TableName(), key)
                )
                if (
                    table.TableName() not in tables
                    and table.TableName() != cls.TableName()
                ):
                    tables.append(table.TableName())
                searchcondition += "`%s`.`%s` %s" % (
                    table.TableName(),
                    columndata[1],
                    condition,
                )
            else:
                searchcondition += "`%s`.`%s` %s" % (cls.TableName(), column, condition)
        searchcondition = "(%s)" % searchcondition[:-4]
        conditions.append(searchcondition)
        return tables, conditions



import os
from typing import Iterable

import urllib.parse
from uweb3.templateparser import Parser


def get_parser():
    return Parser(path=os.path.join(os.path.dirname(__file__), "templates"))


def build_url(table, keys):
    query_args = {}

    for key in keys:
        default = None
        needle = key

        if isinstance(key, tuple):
            needle = key[0]
            default = key[1]

        value = getattr(table, needle)

        if value:
            query_args[needle] = value
        elif default:
            query_args[needle] = default
    return urllib.parse.urlencode(query_args)


class Element:
    def __init__(
        self,
        name,
        value=None,
        attrs=None,
        children: Iterable["Element"] | None = None,
    ):
        self.name = name
        self.value = value
        self.attrs = attrs
        self.children = children

    @property
    def render(self):
        return get_parser().Parse(
            "element.html",
            element=self,
        )


class TableHead(Element):
    def __init__(
        self,
        table,
        col,
        children: Iterable[Element] | None = None,
    ):
        super().__init__(
            name="th",
            value=col.name,
            children=children,
        )

        url = build_url(
            table,
            (
                ("page", 1),
                "query",
            ),
        )
        self.url = f"?{url}&sort_by={col.attr}"
        self.attr = str(col.attr).lower() if col.attr else None
        self.sort_by = table.sort_by.lower() if table.sort_by else None
        self.sort_direction = (
            table.sort_direction.lower() if table.sort_direction else None
        )
        self.sortable = col.sortable

    @property
    def render(self):
        return get_parser().Parse(
            "sortable_head.html",
            element=self,
        )


class TablePagination:
    def __init__(self, table):
        self.sort_url = "&" + build_url(
            table,
            (
                "sort_by",
                "sort_direction",
                "query",
            ),
        )

        if table.page:
            self.page = table.page
        else:
            self.page = 1

        self.previous_page = max(1, self.page - 1)
        self.next_page = self.page + 1

        if table.total_pages:
            self.total_pages = table.total_pages
        else:
            self.total_pages = 0
        self.sliding_range = self._sliding_range()

    @property
    def render(self):
        return get_parser().Parse("pagination.html", element=self)

    def _sliding_range(self):
        nav_end = min(self.previous_page + 4, self.total_pages + 1)
        if self.total_pages - self.page < 2:
            return range(max(1, self.previous_page - 2), nav_end)
        return range(self.previous_page, nav_end)


class SearchField:
    def __init__(self, table):
        self.table = table

    @property
    def render(self):
        return get_parser().Parse("search_field.html", element=self.table)


class TableHeader:
    def __init__(self, table):
        self.table = table

    @property
    def render(self):
        return Element(
            "thead",
            children=[Element("tr", children=self._get_ths())],
        ).render

    def _get_ths(self) -> list:
        return [TableHead(self.table, col) for col in self.table._get_columns()]


class TableBody:
    def __init__(self, table):
        self.table = table

    @property
    def render(self):
        return Element(
            "tbody",
            children=[
                Element(
                    "tr",
                    children=[col.render(item) for col in self.table._get_columns()],
                )
                for item in self.table.items
            ],
        ).render


class Table:
    def __init__(self, children):
        self.children = children
        self.initialized_children = []

    def __call__(self, table):
        self.initialized_children = [child(table) for child in self.children]
        return self

    @property
    def render(self):
        return Element("table", children=self.initialized_children).render



from typing import Optional

import uweb3

from typing import Type
from uweb3.libs.sqltalk.mysql.connection import Connection
from uweb3plugins.core.paginators import table


class SearchableTableMixin:
    @classmethod
    def IntergratedTable(
        cls: Type[uweb3.model.BaseRecord],  # type: ignore
        connection: Connection,
        request_data: uweb3.request.IndexedFieldStorage,
        page_size: int,
        conditions: Optional[list] = None,
        searchable: Optional[list | tuple] = None,
        default_sort: Optional[list[tuple[str, bool]] | None] = None,
    ):

        order = request_data.getfirst("sort_by", None)
        direction = request_data.getfirst("sort_direction", "ASC")
        order_asc = True if direction == "ASC" else False
        page = table.get_current_page(request_data)
        query = request_data.getfirst("query", None)

        if not conditions:
            conditions = []

        if query and searchable:
            conditions.append(
                " OR ".join(
                    [
                        f"{connection.EscapeField(name)} LIKE {connection.EscapeValues(f'%{query}%')}"
                        for name in searchable
                    ]
                )
            )

        data = {
            "offset": max(0, page_size * (page - 1)),
            "limit": page_size,
            "conditions": conditions,
            "yield_unlimited_total_first": True,
        }

        if order:
            data["order"] = [(order, order_asc)]
        if not order and default_sort:
            data["order"] = default_sort

        try:
            results: list[uweb3.model.BaseRecord] = list(
                cls.List(
                    connection,
                    **{key: value for key, value in data.items() if value is not None},
                )
            )
        except connection.OperationalError as exc:
            if exc.args[0] == 1054:
                # Unknown column, this should probably be logged.
                pass
            results = []

        try:
            total_items = int(results.pop(0))
        except (IndexError, ValueError):
            total_items = 0

        return results, total_items, page



import os
from setuptools import setup, find_packages

VERSION = "0.0.4"


def Description():
    """Returns the contents of the README.md file as description information."""
    with open(os.path.join(os.path.dirname(__file__), "README.md")) as r_file:
        return r_file.read()


setup(
    name="uwebthreeplugins",
    version=VERSION,
    description="uWeb3 python3.10 plugin package",
    long_description=Description(),
    long_description_content_type="text/markdown",
    license="ISC",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Environment :: Web Environment",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    author="Stef van Houten",
    author_email="stef@underdark.nl",
    url="https://github.com/stefvanhouten/uweb3plugins",
    keywords="uweb3 plugins",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.10",
)



import uweb3plugins.core.paginators.columns as columns


def _single_get(item, key):
    try:
        val = item[key]
    except (KeyError, TypeError):
        val = getattr(item, key)

    try:
        return val()
    except TypeError:
        return val


def _recursive_getattr(item, keys):
    try:
        keys = keys.split(".")
    except AttributeError:
        pass

    if item is None:
        return None
    if len(keys) == 1:
        return _single_get(item, keys[0])
    else:
        return _recursive_getattr(_single_get(item, keys[0]), keys[1:])


def get_attr(item, attr):
    if isinstance(attr, columns.ConstantAttr):
        return attr.attr

    if "." in attr:
        return _recursive_getattr(item, attr)
    else:
        return _single_get(item, attr)
